package sptossg;

import java.util.*;
import gnu.regexp.*;

/**
 * Translator which takes attribute strings and values from the HTML report and
 * stores them in the appropriate way into a CompanyInfo object
 */
public class AttributeTranslator
{
	CompanyInfo info;	// where data will be stored
	
	// hashtable of translations
	static Hashtable table;	
	static String[] keys = { "Dividend Rate/Share", "Beta", "Shareholders", "Inst. holdings", 
		"Tang. Bk. Value/Share", "Shs. outstg.", "Market cap.", 
		"Avg. daily vol.", "S&P EPS Est.", "P/E on S&P Est." };
	static String[] values = { "dividend rate/share", "beta", "shareholders", "institutional holdings", 
		"book value/share", "shares outstanding (millions)", "market cap (millions)", 
		"average daily volume (millions)", "s&p annual eps estimate", "p/e on s&p annual eps estimate" };
/**
 * This method was created in VisualAge.
 * @param data CompanyInfo object which will receive stored values
 */
public AttributeTranslator(CompanyInfo data)
{
	info = data;
	// initalize the hash
	table = new Hashtable();
	for (int i = 0; i < keys.length; i++) 
		table.put(keys[i], values[i]);
}
/**
 * Add data to the CompanyInfo object
 * @param val value of an attribute - numbers might be re-scaled 
 * @param attr the attribute name, which will be translated before storing
 *    if found in the translation table
 */
public void addData(String val, String attr) throws REException
{
	String newAttr; // the translated attribute name
	String newVal = val.replace('$', ' ');	// don't change the one passed in to us
	if (newVal.equalsIgnoreCase("Nil"))
		newVal = "0";
	

	int pos; // for string searches

	// if it has a year at the end, treat as a periodic attribute		
	RE regex = new RE("(.+) (\\d{4})$");
	REMatch match = regex.getMatch(attr);
	if (match != null)
	{
		addData(newVal, match.toString(1), match.toString(2));
		return;
	}

	// if it contains (M), fix the name
	if ((pos = attr.indexOf("(M)")) != -1)
	{
		// chop off the end, so the translator will recognize it
		newAttr = attr.substring(0, pos - 1);	// skip the space as well
		if (table.containsKey(newAttr))
			newAttr = (String)table.get(newAttr);
		info.set(newVal, newAttr);
		return;
	}
	
	// if it contains (B), translate number to millions and fix the name
	if ((pos = attr.indexOf("(B)")) != -1)
	{
		// chop off the end, so the translator will recognize it
		newAttr = attr.substring(0, pos - 1);	// skip the space as well
		if (table.containsKey(newAttr))
			newAttr = (String)table.get(newAttr);

		// re-scale the value to millions
		double d = Double.valueOf(newVal).doubleValue() * 1000;
		info.set(new Double(d).toString(), newAttr);
		return;
	}

	// default case
	newAttr = (String)table.get(attr);	// the translated attribute name
	if (newAttr == null)
		newAttr = attr;
	info.set(newVal, newAttr);

}
/**
 * Add data to the CompanyInfo object
 * @param val value of an attribute - numbers might be re-scaled 
 * @param attr the attribute name, which will be translated before storing
 *    if found in the translation table
 * @param year year for this attribute
 */
public void addData(String val, String attr, String year) 
{
	String newAttr = (String)table.get(attr);	// the translated attribute name
	if (newAttr == null)
		newAttr = attr;
	info.set(val, newAttr, year);
}
/**
 * main() implementation for testing.
 * @param args An array of attribute strings and values to translate
 * in the form value|attr|[optional year]
 */
public static void main(java.lang.String[] args) throws REException
{
	CompanyInfo data = new CompanyInfo();
	for (int i = 0; i < args.length; i++) 
	{

		AttributeTranslator attrTrans = new AttributeTranslator(data);
		StringTokenizer st = new StringTokenizer(args[i], "|");
		if (st.countTokens() == 2)
		{
			String val = st.nextToken();
			String attr = st.nextToken();
			attrTrans.addData(val, attr);			
		}
		else	// there's some year info present
		{
			String val = st.nextToken();
			String attr = st.nextToken();
			String year = st.nextToken();
			attrTrans.addData(val, attr, year);
		}

	}
	data.dump();	
}
}
