
// Application CommandLoop

package sptossg ;

import java.util.*;
import java.io.*;

public class CommandLoop {

    // MAIN - a routine for testing CompanyInfo storage and retrieval
    public static void main (String args[]) {
	new CommandLoop();
    }

    // I/O routines for the command loop
    String prompt(String pstr) {
	return prompt(pstr, null);
    }

    String attrname(String attr) {
	while (attr.startsWith("[")) {
	    attr = attr.substring(1);
	}
	while (attr.endsWith("]")) {
	    attr = attr.substring(0, attr.length()-1);
	}
	return attr;
    }

    // I/O routines for the command loop

    //   Get a value from the user, using the passed in POSSVAL if there is
    //   one, but if not, then prompt the user for a value instead; A
    //   POSSVAL of "*" indicates a null value without prompting.
    String prompt(String pstr, String possval) {
	if (possval != null) {
	    possval = possval.trim();
	    if (possval.equals("*")) { return ""; }
	    if (! possval.equals("")) { return possval; }
	}

	byte linebytes[] = new byte[1024];
	if (pstr != null) {
	    System.out.print(pstr + ": ");
	    System.out.flush();
	}
	try {
	    System.in.read(linebytes);
	} catch (Exception exc) {
	    System.out.println("exception on read ");

	    return null;
	}

	String line = new String(linebytes);
	line = line.trim();
	if (line.equals("*")) { return ""; }
	return line;
    }


    // The following two functions, wordsep & getWord are used to parse items
    //   The idiom to pull a word off of a line is:
    // 		int pos = wordsep(line);
    // 		String word = getWord(line, pos);
    // 		line = line.substring(pos).trim();

    // separarate words by whitespace
    int wordsep(String line) {
	int last = line.length();

	if (last == 0) { return last; }

        // if the next word starts with '[', then ignore the whitespace and
        // scan till the ']' (but no nesting allowed)
	if (line.trim().charAt(0) == '[') {
	    int close = line.indexOf(']');
	    if (close >= 0 && close <= last) {
		last = close + 1;
	    }
	} else {
	    line = line + " \t";
	    last = Math.min(last,line.indexOf(' '));
	    last = Math.min(last,line.indexOf('\t'));
	}

	return last;
    }

    String getWord(String line, int sep) {
	String itm = line.substring(0,sep);
	itm = itm.trim();

	if (itm.equals("")) { return null; }
	if (itm.charAt(0) == '[') {
	    itm += "]";
	    int end = itm.indexOf(']');
	    itm = itm.substring(1, end);
	}

	return itm;
    }


    // cmdloop - A simple command processor for testing
    public CommandLoop() {
	do_cmds();
    }

    void do_help(String cmd, String params) {
	System.out.println("Available commands: ");
	System.out.println("File I/O routines");
	System.out.println("  parse - Parse an HTML file");
	System.out.println("  convert - Parse an HTML file, writing the " +
			   "result out to a .SSG file");
	System.out.println("  read - read a .SSG file");
	System.out.println("  write - write a .SSG file");
	System.out.println("  load - load internal definitions " +
			   "and dynamic datasets.");
	System.out.println("  save - save internal definitions " +
			   "and dynamic datasets.");
	System.out.println("Internal company manipulation");
	System.out.println("  list - List available companies");
	System.out.println("  rename - Rename a company");
	System.out.println("  use - Switch the current company");
	System.out.println("Data definition and examination");
	System.out.println("  def[ine] - Define a DynamicDataSet");
	System.out.println("  retr[ieve] - retrieve an attribute");
	System.out.println("  set - set an attribute");
	System.out.println("  attr - list the attributes");
	System.out.println("  dump - Display one or all attributes");
	System.out.println("  eval - evaluate a formula expression.");
	System.out.println("Other Commands");
	System.out.println("  echo - echo a line (for scripting)");
	System.out.println("  prompt - turn prompt on/off (for scripting)");
	System.out.println("  quit - exit the program");
    }

    void do_define(String cmd, String params) {

	int pos = wordsep(params);
	String word = getWord(params, pos);
	params = params.substring(pos).trim();

	String attr = prompt("Attribute Name", word);
	String expr = prompt("Attribute Expression", params);
	attr = attrname(attr);
	DynamicDataSet dds = DynamicDataSet.define(attr, expr);
	if (dds == null) {
	    System.out.println("Parse error defining " + attr);
	} else {
	    System.out.println("Defined dynamic attribute: " + dds);
	}
    }

    void do_setup(String cmd, String params) {

	String attrs[][] =
	{ {"PTP Margin", "100 * [PreTax Inc] / Sales"},
	  {"Sales Increase", "100 * Sales / prev(Sales)"}
	};

	for (int i = 0; i < attrs.length; i++) {
	    String aname = attrs[i][0];
	    String aformula = attrs[i][1];

	    DynamicDataSet.define(aname, aformula);
	    System.out.println("Defined dynamic attribute: " +
			       aname + " => " + aformula);
	}
    }

    void do_parse(String cmd, String params) {
	String coname = prompt("Company", params);

	curcompany = new CompanyInfo();
	allcompanies.put(coname, curcompany);

	try {
	    SPReportParser parser = new SPReportParser(coname + ".html");
	    if (!parser.parse(curcompany)) {
		System.err.println("parse() failed with the message: "
				   + parser.parseError());
	    } else {
		System.out.println(coname + ".html parsed. ");
	    }
	} catch (Exception exc) {
	    System.out.println("Error: " + exc);
	}
    }

    class fileMatcher implements FilenameFilter {
	String start;
	String end;

	fileMatcher(String start, String end) {
	    this.start = start;
	    this.end = end;
	}

	public boolean accept(File dir, String name) {
	    if (start != null && (! name.startsWith(start))) { return false; }
	    if (end != null && (! name.endsWith(end))) { return false; }
	    return true;
	}
    }

    void do_convert(String cmd, String params) {

	String coname = prompt("Company", params);

	String[] convlist;

	// Expand if directory
	File file = new File(coname);
	if (file.isDirectory()) {
	    convlist = file.list(new fileMatcher(null, ".html"));
	    for (int i = 0; i < convlist.length; i++) {
		String fname = new File(coname, convlist[i]).getPath();
		convlist[i] = fname;
	    }
	} else {
	    convlist = new String[1];
	    convlist[0] = coname + ".html";
	}

	for (int i = 0; i < convlist.length; i++) {
	    coname = convlist[i];
	    String hfile = coname;
	    coname = coname.substring(0,coname.length()-5);
	    String sfile = coname + ".ssg";

	    boolean write_file = true;

	    // Read and parse HTML file
	    CompanyInfo company = new CompanyInfo();
	    try {
		SPReportParser parser = new SPReportParser(hfile);
		if (!parser.parse(company)) {
		    System.err.println("parse() on \"" + coname +
				       "\" failed with the message: "
				       + parser.parseError());
		    write_file = false;
		} else {
		    System.out.println(coname + ".html parsed. ");
		}
	    } catch (Exception exc) {
		System.out.println("Error: " + exc);
		write_file = false;
	    }

	    if (write_file) {
		// Setup attributes processing needing handling
		process_attrs(company);

		// Write out SSG file
		company.writeSSGfile(sfile);
		System.out.println(sfile + " written. ");

		// Clean up attribute group modifications
		finish_attrs(company);

		// **************** remove these to deactivate
		// This preserves conversion info (company & DynamicDataSets) for
		// debugging.
		curcompany = company;
		DynamicDataSet.activate("converter", true);
	    }
	}
    }

    String curgroup;


    void process_periodic_attrs(CompanyInfo company) {
	// Name translations for periodic attributes
	// Translation process will use formulas
	String periodic_attrs[][] =
	{
	    {"Quarterly E/S", "[quarterly eps]"},
	    {"Quarterly Sales", "[quarterly revenues]"},
	    {"High Price", "[Price - High]"},
	    {"Low Price", "[Price - Low]"},
	    {"Book Value", "[Tangible Bk. Val.]"},
	    {"Tax Rate", "100 * [Eff. Tax Rate]"},
	    {"Net Income", "[Net Inc.]"},
	    {"Sales", "[Revs.]"},
	    {"PreTax Inc", "[Pretax Inc.]"},
	    {"Earnings per Share", "[Earnings]"},
	    {"Long Term Debt", "[LT Debt]"},
	    {"Dividend per Share", "[Dividends]"},
	    {"Cash flow/share", "[Cash Flow Per Share]"},
	    {"Last Year of data", "(latest(year(), 1))"},
	    {"Month of Latest Quarterly Data", "(3  * latest(period(), 4))"},
	    {"Year of Latest Quarterly Data", "(latest(year(), 4))"},
	    {"Last CY of data +5", "5 + [Last Year of data]"},

	    {"unused", "item - end"}
	};

	for (int i = 0; i < periodic_attrs.length; i++) {
	    String aname = periodic_attrs[i][0];
	    String aformula = periodic_attrs[i][1];

	    DynamicDataSet.define(aname, aformula);
	}
    }

    void process_static_attrs(CompanyInfo company) {
	// Name translations for static attributes
	// Translation process will copy values
	String static_attrs[][] =
	{
	    {"Current P/E", "12 month p/e"},
	    {"52 Week High Price", "52 week high price"},
	    {"52 Week Low Price", "52 week low price"},
	    {"Beta", "beta"},
	    {"Current Price", "current price"},
	    {"Exchange", "exchange"},
	    {"Industry", "industry"},
	    {"Name", "name"},
	    {"Date of Data Creation", "report date"},
	    {"Symbol", "ticker"},
	    {"Number of Common Shares Outstanding", "shares outstanding (millions)"},
	    {"Industry Group", "Industry"},
	    {"Institutional Holdings", "institutional holdings"},
	    // Following isn't completely right, this will store a
	    // string, e.g. "Dec. 31"
	    {"Fiscal Year End Month", "fy ending"},

//	    {"", ""},

	    {"unused", "item - end"}
	};

	for (int i = 0; i < static_attrs.length; i++) {
	    String aname = static_attrs[i][0];
	    String asource = static_attrs[i][1];

	    String val = company.get(asource);
	    if (val != null) {
		company.set(val, aname);
	    }
	}
    }

    void process_constant_attrs(CompanyInfo company) {
	// Name translations for constant attributes
	// Translation process will store values
	String static_attrs[][] =
	{
	    {"Data Source", "S&P Report"},

	    {"unused", "item - end"}
	};

	for (int i = 0; i < static_attrs.length; i++) {
	    String aname = static_attrs[i][0];
	    String val = static_attrs[i][1];

	    if (val != null) {
		company.set(val, aname);
	    }
	}
    }

    void process_attrs(CompanyInfo company) {
	// save the old group for restoring
	curgroup = DynamicDataSet.curGroup();
	// Create and active a DDS group for conversions
	DynamicDataSet.newGroup("converter", 1000);
	DynamicDataSet.activate("converter");

	process_periodic_attrs(company);
	process_static_attrs(company);
	process_constant_attrs(company);
    }

    // When done with conversions, deactivate DDS group and restore
    // original group as default
    void finish_attrs(CompanyInfo company) {
	DynamicDataSet.activate("converter", false);
	DynamicDataSet.useGroup(curgroup);
    }


    void do_read(String cmd, String params) {
	String coname = prompt("Company", params);

	curcompany = new CompanyInfo();
	allcompanies.put(coname, curcompany);

	curcompany.loadSSGfile(coname + ".ssg");
	System.out.println(coname + ".ssg read. ");
    }

    void do_write(String cmd, String params) {
	if (curcompany == null) {
	    System.out.println("No current company.");
	} else {

	    String ticker = curcompany.get("Symbol");
	    String defstr = "";
	    if (ticker != null) {
		defstr = " (default is " + ticker + ")";
	    }

	    String coname = prompt("Company" + defstr, params);

	    if (coname.equals("")) { coname = ticker; }

	    String fname = coname + ".ssg";

	    curcompany.writeSSGfile(fname);
	    System.out.println(fname + " written. ");
	}
    }

    void do_list(String cmd, String params) {
	Enumeration e;

	int i = 0;
	e = allcompanies.keys();
	while (e.hasMoreElements()) {
	    String key = (String) e.nextElement();
	    System.out.println(i++ + ": " + key);
	}
	if (i == 0) {
	    System.out.println("No companies loaded.");
	}
    }

    void do_rename(String cmd, String params) {
	int sep = wordsep(params);

	String coname = prompt("Company", params.substring(0, sep));
	String newname = prompt("New company id", params.substring(sep));

	CompanyInfo co = (CompanyInfo) allcompanies.get(coname);
	if (co == null) {
	    System.out.println("No company named '" + coname + "'");
	} else {
	    allcompanies.put(newname, co);
	    allcompanies.remove(coname);
	}
    }

    void do_use(String cmd, String params) {
	String coname = prompt("Company", params);

	curcompany = (CompanyInfo) allcompanies.get(coname);
	if (curcompany == null) {
	    System.out.println("No company named '" + coname + "'");
	}
    }

    void do_setretr(String cmd, String params) {

	if (curcompany == null) {
	    System.out.println("No current company.");
	} else {

	    int pos = wordsep(params);
	    String word = getWord(params, pos);
	    params = params.substring(pos).trim();

	    String attr = prompt("Attribute", word);
	    attr = attrname(attr);

	    String val;
	    String perref;

	    Period per;

	    pos = wordsep(params);
	    word = getWord(params, pos);
	    params = params.substring(pos).trim();

	    String year = prompt("Year", word);
	    if (year.equals("")) {
		per = new Period();
		perref = "";
	    } else {
		pos = wordsep(params);
		word = getWord(params, pos);
		params = params.substring(pos).trim();

		String period = prompt("Period", word);
		if (period.equals("")) {
		    per = new Period(year);
		    perref = "(" + year + ")";
		} else {
		    per = new Period(year, period);
		    perref = "(" + year + "," + period + ")";
		}
	    }

	    if (cmd.equals("set")) {
		val = prompt("Value", params);
		curcompany.set(val, attr, per);
	    } else {
		val = curcompany.get(attr, per);
	    }
	    System.out.println(attr + perref + " = " + val);
	}
    }

    void do_eval(String cmd, String params) {

	if (curcompany == null) {
	    System.out.println("No current company.");
	} else {
	    double val;
	    String perref;

	    Period per;

	    int pos = wordsep(params);
	    String word = getWord(params, pos);
	    params = params.substring(pos).trim();

	    String year = prompt("Year", word);
	    if (year.equals("")) {
		per = new Period();
		perref = "";
	    } else {
		pos = wordsep(params);
		word = getWord(params, pos);
		params = params.substring(pos).trim();

		String period = prompt("Period", word);
		if (period.equals("")) {
		    per = new Period(year);
		    perref = "(" + year + ")";
		} else {
		    per = new Period(year, period);
		    perref = "(" + year + "," + period + ")";
		}
	    }

	    String expr = prompt("Expression", params);

	    try {
		val = Formula.eval(expr, curcompany, per);
		System.out.println("Expression" + perref + " = " + val);
	    } catch (Exception exc) {
		System.out.println("Parse Error: " + exc +
				   "\nOn expression " + expr);
	    }
	}
    }

    void do_attr(String cmd, String params) {
	if (curcompany == null) { System.out.println("No current company.");
	} else { curcompany.dumpAttrs(); }
    }

    void do_dump(String cmd, String params) {
	if (curcompany == null) {
	    System.out.println("No current company.");
	} else {
	    int pos = wordsep(params);
	    String word = getWord(params, pos);
	    params = params.substring(pos).trim();
	    String attr = prompt("Attribute", word);
	    attr = attrname(attr);
	    if (attr.equals("")) {
		try {
		    curcompany.dump();
		} catch (Exception exc) {
		    System.out.println("Error dumping all attributes: " + exc);
		}
	    } else {
		try {
		    curcompany.dump(attr);
		} catch (Exception exc) {
		    System.out.println("Error dumping attribute " +
				       attr + ": " + exc);
		}
	    }
	}
    }

    void do_dprint(String cmd, String params) {
	if (curcompany == null) {
	    System.out.println("No current company.");
	} else {
	    int pos = wordsep(params);
	    String word = getWord(params, pos);
	    params = params.substring(pos).trim();
	    String attr = prompt("Attribute", word);
	    attr = attrname(attr);
	    if (attr.equals("")) {
		try {
		    curcompany.dprint();
		} catch (Exception exc) {
		    System.out.println("Error dprinting company: " + exc);
		}
	    } else {
		try {
		    curcompany.dprint(attr);
		} catch (Exception exc) {
		    System.out.println("Error dprinting attribute " +
				       attr + ": " + exc);
		}}}}

    void do_test(String cmd, String params) {
	String defaultval = "/tmp";

	String fname = prompt("File", params);

	if (fname.equals("")) { fname = defaultval; }

	File file = new File(fname);
	if (file.isDirectory()) {
	    System.out.println(fname + " is a directory.  HTML files in it are: ");
	    String[] hfiles = file.list(new fileMatcher(null, ".html"));
	    for (int i = 0; i < hfiles.length; i++) {
		System.out.println(i + ": " + hfiles[i]);
	    }
	} else {
	    System.out.println(fname + " is a file.");
	}
    }

    Hashtable allcompanies;
    CompanyInfo curcompany;

    void do_cmds () {
	curcompany = null;
	allcompanies = new Hashtable();

	System.out.println("\nSSG Program Command Line processor");
	System.out.println("Type 'help' for a list of commands\n");

	boolean prompt = true;

	while (true) {
	    String line;

	    line = prompt(prompt ? "Command" : null);

	    // A bracketed attribute is a shortcut for "dump [attr]"
	    if (line.startsWith("[")) { line = "dump " + line; }

	    int sep = wordsep(line);

	    String cmd = line.substring(0,sep).toLowerCase();
	    String params = line.substring(sep).trim();

	    // Exit the command loop
	    if (cmd.equals("quit")) { return; }
	    // list commands
	    else if (cmd.equals("help")) { do_help(cmd, params); }
	    // echo a line
	    else if (cmd.equals("echo")) { System.out.println(params); }
	    // toggle prompting
	    else if (cmd.equals("prompt")) {
		prompt = ! prompt;
		System.out.println("Prompting is " + (prompt ? "on" : "off"));
	    }
	    // Define a DynamicDataSet
	    else if (cmd.equals("def") || cmd.equals("define")) {
		do_define(cmd, params); }
	    // Initialize some formulas
	    else if (cmd.equals("setup")) { do_setup(cmd, params); }
	    // toggle debugging
	    else if (cmd.equals("debug")) { Debug.toggle(); }
	    // Parse an HTML file
	    else if (cmd.equals("parse")) { do_parse(cmd, params); }
	    // Convert an HTML file to a .SSG file
	    else if (cmd.equals("conv") || cmd.equals("convert")) {
		do_convert(cmd, params); }
	    // read data from a .SSG file
	    else if (cmd.equals("read")) { do_read(cmd, params); }
	    // write the current company out as a .SSG file
	    else if (cmd.equals("write")) { do_write(cmd, params); }
	    // save or load global settings and dynamic attributes
	    else if (cmd.equals("save") || cmd.equals("load")) { // !!!! TBD !!!!
		System.out.println("Can't " + cmd + " definitions yet. ");
	    }
	    // List available companies
	    else if (cmd.equals("list")) { do_list(cmd, params); }
	    // Rename a company
	    else if (cmd.equals("rename")) { do_rename(cmd, params); }
	    // Switch the current company
	    else if (cmd.equals("use")) { do_use(cmd, params); }
	    // retrieve or set an attribute in the current company
	    else if (cmd.equals("retr") || cmd.equals("retrieve") ||
		cmd.equals("set")) { do_setretr(cmd, params); }
	    // evaluate a formula in the current company
	    else if (cmd.equals("eval") || cmd.equals("evaluate")) {
		do_eval(cmd, params); }
	    // list the attributes of the current company
	    else if (cmd.equals("attr")) { do_attr(cmd, params); }
	    // Display the current company
	    else if (cmd.equals("dump")) { do_dump(cmd, params); }
	    // Display debugging contents of objects
	    else if (cmd.equals("dprint")) { do_dprint(cmd, params); }
	    // testing routine (unpublished)
	    else if (cmd.equals("test")) { do_test(cmd, params); }
}}
}
